@@ -10,6 +10,8 @@ import getpass
 import functools
 import itertools
 import distutils.version
+import hashlib
+import tempfile
 
 import requests
 import requests.auth
@@ -61,7 +63,7 @@ class Session(object):
     def __init__(
         self, server_url=None, api_key=None, api_user=None, auto_populate=True,
         plugin_paths=None, cache=None, cache_key_maker=None,
-        auto_connect_event_hub=True
+        auto_connect_event_hub=True, schema_cache_path=None
     ):
         '''Initialise session.
 
@@ -92,14 +94,14 @@ class Session(object):
 
             The session will add the specified cache to a pre-configured layered
             cache that specifies the top level cache as a
-            :class:`ftrack_api.cache.MemoryCache`. Therefore, it is unnecessary 
+            :class:`ftrack_api.cache.MemoryCache`. Therefore, it is unnecessary
             to construct a separate memory cache for typical behaviour. Working
             around this behaviour or removing the memory cache can lead to
             unexpected behaviour.
 
         *cache_key_maker* should be an instance of a key maker that fulfils the
-        :class:`ftrack_api.cache.KeyMaker` interface and will be used to 
-        generate keys for objects being stored in the *cache*. If not specified, 
+        :class:`ftrack_api.cache.KeyMaker` interface and will be used to
+        generate keys for objects being stored in the *cache*. If not specified,
         a :class:`~ftrack_api.cache.StringKeyMaker` will be used.
 
         If *auto_connect_event_hub* is True then embedded event hub will be
@@ -109,6 +111,12 @@ class Session(object):
         manually connected using :meth:`EventHub.connect
         <ftrack_api.event.hub.EventHub.connect>`.
 
+        Enable schema caching by setting *schema_cache_path* to a folder path.
+        If not set, :envvar:`FTRACK_API_SCHEMA_CACHE_PATH` will be used to
+        determine the path to store cache in. If the environment variable is
+        also not specified then a temporary directory will be used. Set to
+        `False` to disable schema caching entirely.
+
         '''
         super(Session, self).__init__()
         self.logger = logging.getLogger(
@@ -178,8 +186,6 @@ class Session(object):
 
             self.cache.caches.append(cache)
 
-        self._attached = collections.OrderedDict()
-
         self._request = requests.Session()
         self._request.auth = SessionAuthentication(
             self._api_key, self._api_user
@@ -213,7 +219,17 @@ class Session(object):
 
         # TODO: Make schemas read-only and non-mutable (or at least without
         # rebuilding types)?
-        self.schemas = self._fetch_schemas()
+        if schema_cache_path is not False:
+            if schema_cache_path is None:
+                schema_cache_path = os.environ.get(
+                    'FTRACK_API_SCHEMA_CACHE_PATH', tempfile.gettempdir()
+                )
+
+            schema_cache_path = os.path.join(
+                schema_cache_path, 'ftrack_api_schema_cache.json'
+            )
+
+        self.schemas = self._load_schemas(schema_cache_path)
         self.types = self._build_entity_type_classes(self.schemas)
 
         self._configure_locations()
@@ -243,6 +259,11 @@ class Session(object):
         '''Return event hub.'''
         return self._event_hub
 
+    @property
+    def _local_cache(self):
+        '''Return top level memory cache.'''
+        return self.cache.caches[0]
+
     def check_server_compatibility(self):
         '''Check compatibility with connected server.'''
         server_version = self.server_information.get('version')
@@ -269,29 +290,37 @@ class Session(object):
                 )
 
     def reset(self):
-        '''Reset session clearing all locally stored data.
+        '''Reset session clearing local state.
 
-        If the cache used by the session is a
+        Clear all pending operations and expunge all entities from session.
+
+        Also clear the local cache. If the cache used by the session is a
         :class:`~ftrack_api.cache.LayeredCache` then only clear top level cache.
         Otherwise, clear the entire cache.
 
+        Plugins are not rediscovered or reinitialised, but certain plugin events
+        are re-emitted to properly configure session aspects that are dependant
+        on cache (such as location plugins).
+
+        .. warning::
+
+            Previously attached entities are not reset in memory and will retain
+            their state, but should not be used. Doing so will cause errors.
+
         '''
         if self.recorded_operations:
             self.logger.warning(
                 'Resetting session with pending operations not persisted.'
             )
 
-        if isinstance(self.cache, ftrack_api.cache.LayeredCache):
-            try:
-                self.cache.caches[0].clear()
-            except IndexError:
-                pass
-        elif isinstance(self.cache, ftrack_api.cache.Cache):
-            self.cache.clear()
-
-        self._attached.clear()
+        # Clear pending operations.
         self.recorded_operations.clear()
-        self._request.close()
+
+        # Clear top level cache (expected to be enforced memory cache).
+        self._local_cache.clear()
+
+        # Re-configure certain session aspects that may be dependant on cache.
+        self._configure_locations()
 
     def auto_populating(self, auto_populate):
         '''Temporarily set auto populate to *auto_populate*.
@@ -322,7 +351,7 @@ class Session(object):
     @property
     def created(self):
         '''Return list of newly created entities.'''
-        entities = self._attached.values()
+        entities = self._local_cache.values()
         states = ftrack_api.inspection.states(entities)
 
         return [
@@ -333,7 +362,7 @@ class Session(object):
     @property
     def modified(self):
         '''Return list of locally modified entities.'''
-        entities = self._attached.values()
+        entities = self._local_cache.values()
         states = ftrack_api.inspection.states(entities)
 
         return [
@@ -344,7 +373,7 @@ class Session(object):
     @property
     def deleted(self):
         '''Return list of deleted entities.'''
-        entities = self._attached.values()
+        entities = self._local_cache.values()
         states = ftrack_api.inspection.states(entities)
 
         return [
@@ -657,25 +686,6 @@ class Session(object):
 
         return data
 
-    def _attach(self, entity):
-        '''Attach *entity* to session if not already.'''
-        key = str(ftrack_api.inspection.identity(entity))
-        current = self._attached.get(key)
-
-        if current is None:
-            self._attached[key] = entity
-
-        elif current is not entity:
-            raise ValueError(
-                'Cannot attach {0!r}. A different instance {1!r} of that '
-                'entity is already attached.'.format(entity, current)
-            )
-
-    def _detach(self, entity):
-        '''Detach *entity* from session.'''
-        key = str(ftrack_api.inspection.identity(entity))
-        del self._attached[key]
-
     def merge(self, value, merged=None):
         '''Merge *value* into session and return merged value.
 
@@ -756,14 +766,19 @@ class Session(object):
                     .format(entity_key, attached_entity, id(attached_entity))
                 )
                 return attached_entity
+            else:
+                self.logger.debug(
+                    'Entity not already processed for key {0}. Keys: {1}'
+                    .format(entity_key, sorted(merged.keys()))
+                )
 
             # Check for existing instance of entity in cache.
             self.logger.debug(
                 'Checking for entity in cache with key {0}'.format(entity_key)
             )
-
             try:
                 attached_entity = self.cache.get(entity_key)
+                from_cache = True
                 self.logger.debug(
                     'Retrieved existing entity from cache: {0} at {1}'
                     .format(attached_entity, id(attached_entity))
@@ -774,6 +789,7 @@ class Session(object):
                 attached_entity = self._create(
                     entity.entity_type, {}, reconstructing=True
                 )
+                from_cache = False
                 self.logger.debug(
                     'Entity not present in cache. Constructed new instance: '
                     '{0} at {1}'.format(attached_entity, id(attached_entity))
@@ -782,12 +798,14 @@ class Session(object):
             # Mark entity as seen to avoid infinite loops.
             merged[entity_key] = attached_entity
 
-            # Expand references. This is required as a serialised cache might
-            # have returned just a plain entity object with the rest of the data
-            # stored separately. The reason this is done here rather in the
-            # specific cache is so that any higher level cache can be taken
-            # advantage of when fetching data.
-            self._merge_references(attached_entity, merged=merged)
+            # Expand references when entity instance retrieved from cache.
+            # This is required as a serialised cache might have returned an
+            # entity instance that has references to other entities not yet
+            # merged. The reason this is done here rather in the specific cache
+            # is so that any higher level cache can be taken advantage of when
+            # fetching data for referenced entities.
+            if from_cache:
+                self._merge_references(attached_entity, merged=merged)
 
             # Merge new entity data into cache entity. If this causes the cache
             # entity to change then persist those changes back to the cache.
@@ -802,13 +820,28 @@ class Session(object):
                     'detected.'
                 )
 
-            # Ensure this instance is now attached to the session.
-            self._attach(attached_entity)
-
         return attached_entity
 
     def _merge_references(self, entity, merged=None):
         '''Recursively merge entity references in *entity*.'''
+        # As optimisation, mark inflated entities and avoid inflating more than
+        # once. This is possible because *entity* should always be the same
+        # entity instance retrieved from the top level memory cache.
+        # TODO: Consider refactor API encoding to explicitly mark references
+        # as such, perhaps by a private attribute __is_reference__ in order to
+        # make expansion of references easier to determine and on a
+        # per-reference basis.
+        is_inflated = getattr(entity, '_inflated', False)
+        if is_inflated:
+            self.logger.debug(
+                'Skipping merging references as entity appears to already have '
+                'been inflated.'
+            )
+            return
+
+        else:
+            entity._inflated = True
+
         self.logger.debug('Merging references.')
 
         if merged is None:
@@ -821,7 +854,7 @@ class Session(object):
             if isinstance(
                 local_value,
                 (
-                    ftrack_api.entity.base.Entity, 
+                    ftrack_api.entity.base.Entity,
                     ftrack_api.collection.Collection,
                     ftrack_api.collection.MappedCollectionProxy
                 )
@@ -840,7 +873,7 @@ class Session(object):
             if isinstance(
                 remote_value,
                 (
-                    ftrack_api.entity.base.Entity, 
+                    ftrack_api.entity.base.Entity,
                     ftrack_api.collection.Collection,
                     ftrack_api.collection.MappedCollectionProxy
                 )
@@ -1111,25 +1144,18 @@ class Session(object):
         if batch:
             result = self._call(batch)
 
-            # Clear all local values for committed attributes before proceeding
-            # with merge. Otherwise it is possible for an immutable attribute
-            # error to be bypassed.
-            with self.operation_recording(False):
-                for payload in batch:
-                    if payload['action'] in ('create', 'update'):
-                        # Retrieve entity from cache.
-                        entity = self._get(
-                            payload['entity_type'], payload['entity_key']
-                        )
+            # Clear recorded operations.
+            self.recorded_operations.clear()
 
-                        for key in payload['entity_data'].keys():
-                            if key in ('__entity_type__', ):
-                                continue
-
-                            attribute = entity.attributes.get(key)
-                            attribute.set_local_value(
-                                entity, ftrack_api.symbol.NOT_SET
-                            )
+            # As optimisation, clear local values which are not primary keys to
+            # avoid redundant merges when merging references. Note: primary keys
+            # remain as needed for cache retrieval on new entities.
+            with self.auto_populating(False):
+                with self.operation_recording(False):
+                    for entity in self._local_cache.values():
+                        for attribute in entity:
+                            if attribute not in entity.primary_key_attributes:
+                                del entity[attribute]
 
             # Process results merging into cache relevant data.
             for entry in result:
@@ -1143,8 +1169,52 @@ class Session(object):
                     # TODO: Expunge entity from cache.
                     pass
 
-            # Clear operations.
-            self.recorded_operations.clear()
+            # Clear remaining local state, including local values for primary
+            # keys on entities that were merged.
+            with self.auto_populating(False):
+                with self.operation_recording(False):
+                    for entity in self._local_cache.values():
+                        entity.clear()
+
+    def rollback(self):
+        '''Clear all recorded operations and local state.
+
+        Typically this would be used following a failed :meth:`commit` in order
+        to revert the session to a known good state.
+
+        Newly created entities not yet persisted will be detached from the
+        session / purged from cache and no longer contribute, but the actual
+        objects are not deleted from memory. They should no longer be used and
+        doing so could cause errors.
+
+        '''
+        with self.auto_populating(False):
+            with self.operation_recording(False):
+
+                # Detach all newly created entities and remove from cache. This
+                # is done because simply clearing the local values of newly
+                # created entities would result in entities with no identity as
+                # primary key was local while not persisted. In addition, it
+                # makes no sense for failed created entities to exist in session
+                # or cache.
+                for operation in self.recorded_operations:
+                    if isinstance(
+                        operation, ftrack_api.operation.CreateEntityOperation
+                    ):
+                        entity_key = str((
+                            str(operation.entity_type),
+                            operation.entity_key.values()
+                        ))
+                        try:
+                            self.cache.remove(entity_key)
+                        except KeyError:
+                            pass
+
+                # Clear locally stored modifications on remaining entities.
+                for entity in self._local_cache.values():
+                    entity.clear()
+
+        self.recorded_operations.clear()
 
     def _fetch_server_information(self):
         '''Return server information.'''
@@ -1167,10 +1237,105 @@ class Session(object):
         '''
         ftrack_api.plugin.discover(self._plugin_paths, [self])
 
-    def _fetch_schemas(self):
-        '''Return schemas fetched from server.'''
-        result = self._call([{'action': 'query_schemas'}])
-        return result[0]
+    def _read_schemas_from_cache(self, schema_cache_path):
+        '''Return schemas and schema hash from *schema_cache_path*.
+
+        *schema_cache_path* should be the path to the file containing the
+        schemas in JSON format.
+
+        '''
+        self.logger.debug(
+            'Reading schemas from cache {0!r}'.format(schema_cache_path)
+        )
+
+        if not os.path.exists(schema_cache_path):
+            self.logger.info(
+                'Cache file not found at {0!r}.'.format(
+                    schema_cache_path
+                )
+            )
+
+            return [], None
+
+        with open(schema_cache_path, 'r') as schema_file:
+            schemas = json.load(schema_file)
+            hash_ = hashlib.md5(
+                json.dumps(schemas, sort_keys=True)
+            ).hexdigest()
+
+        return schemas, hash_
+
+    def _write_schemas_to_cache(self, schemas, schema_cache_path):
+        '''Write *schemas* to *schema_cache_path*.
+
+        *schema_cache_path* should be a path to a file that the schemas can be
+        written to in JSON format.
+
+        '''
+        self.logger.debug(
+            'Updating schema cache {0!r} with new schemas.'
+            .format(schema_cache_path)
+        )
+
+        with open(schema_cache_path, 'w') as local_cache_file:
+            json.dump(schemas, local_cache_file, indent=4)
+
+    def _load_schemas(self, schema_cache_path):
+        '''Load schemas.
+
+        First try to load schemas from cache at *schema_cache_path*. If the
+        cache is not available or the cache appears outdated then load schemas
+        from server and store fresh copy in cache.
+
+        If *schema_cache_path* is set to `False`, always load schemas from
+        server bypassing cache.
+
+        '''
+        local_schema_hash = None
+        schemas = []
+
+        if schema_cache_path:
+            try:
+                schemas, local_schema_hash = self._read_schemas_from_cache(
+                    schema_cache_path
+                )
+            except (IOError, TypeError, AttributeError, ValueError):
+                # Catch any known exceptions when trying to read the local
+                # schema cache to prevent API from being unusable.
+                self.logger.exception(
+                    'Schema cache could not be loaded from {0!r}'
+                    .format(schema_cache_path)
+                )
+
+        # Use `dictionary.get` to retrieve hash to support older version of
+        # ftrack server not returning a schema hash.
+        server_hash = self._server_information.get(
+            'schema_hash', False
+        )
+        if local_schema_hash != server_hash:
+            self.logger.debug(
+                'Loading schemas from server due to hash not matching.'
+                'Local: {0!r} != Server: {1!r}'.format(
+                    local_schema_hash, server_hash
+                )
+            )
+            schemas = self._call([{'action': 'query_schemas'}])[0]
+
+            if schema_cache_path:
+                try:
+                    self._write_schemas_to_cache(schemas, schema_cache_path)
+                except (IOError, TypeError):
+                    self.logger.exception(
+                        'Failed to update schema cache {0!r}.'
+                        .format(schema_cache_path)
+                    )
+
+        else:
+            self.logger.debug(
+                'Using cached schemas from {0!r}'.format(schema_cache_path)
+            )
+
+        return schemas
 
     def _build_entity_type_classes(self, schemas):
         '''Build default entity type classes.'''
@@ -1633,7 +1798,7 @@ class Session(object):
             data.setdefault('file_type', os.path.splitext(container_path)[-1])
 
             container = self._create_component(
-                'SequenceComponent', container_path, data, location
+                'SequenceComponent', container_path, data, location=None
             )
 
             # Create member components for sequence.
@@ -1645,8 +1810,17 @@ class Session(object):
                     'file_type': os.path.splitext(member_path)[-1]
                 }
 
-                self._create_component(
-                    'FileComponent', member_path, member_data, location
+                component = self._create_component(
+                    'FileComponent', member_path, member_data, location=None
+                )
+                container['members'].append(component)
+
+            if location:
+                origin_location = self.get(
+                    'Location', ftrack_api.symbol.ORIGIN_LOCATION_ID
+                )
+                location.add_component(
+                    container, origin_location, recursive=True
                 )
 
             return container
